import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace



def perm_init(shape,dtype=None):
  #works only for 16x16
  data=np.zeros((16,16))

  data[0,0]=1
  data[1,2]=1
  data[2,1]=1
  data[3,5]=1
  data[4,4]=1
  data[5,3]=1
  data[6,9]=1
  data[7,8]=1
  data[8,6]=1
  data[9,7]=1
  data[10,12]=1
  data[11,13]=1
  data[12,11]=1
  data[13,14]=1
  data[14,15]=1
  data[15,10]=1
  
  return t.constant(data,dtype=dtype)

class gperm(Layer):#multiplies all elements with one orthogonal matrix
  def __init__(self,gs=20,param=40,**kwargs):
    self.gs=gs
    self.param=param
    super(gperm,self).__init__(input_shape=(gs,param))

  def build(self, input_shape):

    self.trafo=self.add_weight(name='orthogonal_trafo',#Matrix N
                               shape=(self.param,self.param),
                               initializer=perm_init,
                               trainable=False)



    self.built=True


  def call(self,x):
    x=x[0]

    x=K.reshape(x,(-1,self.gs,self.param))


    x=K.dot(x,self.trafo)
    
    return x


  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    if len(input_shape)==2 and self.gs==1:
      assert input_shape[1]==self.param
      return tuple([input_shape[0],self.gs,self.param])
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    return tuple([input_shape[0],self.gs,self.param])    

  def get_config(self):
    mi={"gs":self.gs,"param":self.param}
    th=super(gperm,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gperm(**config)
















import os
from os.path import isfile


with open("__init__.py","w") as f:

  for q in os.listdir("."):
    if not isfile(q):continue
    if "_" in q:continue
    if "makeinit.py" in q:continue
    if not ".py" in q:continue

    q=q.replace(".py","")

    f.write("import grapa.layers."+q+"\n")




print("done")





import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gtbuilder(Layer):
  def __init__(self,gs=30,param=10,free=30,**kwargs):
    self.gs=gs
    self.param=param
    self.free=free
    super(gtbuilder,self).__init__(input_shape=(gs,gs+gs+param))

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"free":self.free}
    th=super(gtbuilder,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gtbuilder(**config)


  def getmetrik(self):
    return [[1],[1]]
  def fromdistsq(self,dsq):
    return K.exp(-dsq)

  def build(self, input_shape):

    self.built=True

  def call(self,x):
    #print(x.shape)
   
    gs=self.gs 
    mata=x[:,:gs,:gs]
    matb=x[:,:gs,gs:gs+gs]
    data=x[:,:gs,gs+gs:]

    mata=K.reshape(mata,(-1,gs,gs,1))
    matb=K.reshape(matb,(-1,gs,gs,1))
     
    mat=K.concatenate((mata,matb),axis=-1)
    metrik=K.constant(self.getmetrik())
    
    dsq=K.reshape(K.dot(mat,metrik),(-1,gs,gs))
    
    basegraph=self.fromdistsq(dsq)
    
    
     
    #print(metrik.shape,xm.shape,xt.shape,dsq.shape)
    #print(basegraph.shape)
    
    parax=x[:,:,self.gs+self.gs:]#please note, that this layer is build to read two times the same data: first the one for the distance generation, and afterwards the one for node data, also note, that glbuilder does not do this, but instead works on the same dataset
    #print(parax.shape)
    
    if self.free==0:return K.concatenate((basegraph,parax),axis=-1) 
    zero1=K.zeros_like(x[:,:,0])
    zero1=K.reshape(zero1,(-1,x.shape[1],1))
    #print("!",zero1.shape)
    zerolis=[]
    for i in range(self.free):
      zerolis.append(zero1)
    zeros=K.concatenate(zerolis,axis=-1)
    #print(zeros.shape)
    
    return K.concatenate((basegraph,parax,zeros))

    
  def compute_output_shape(self,input_shape):
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.gs+self.gs+self.param
    return tuple([input_shape[0],self.gs,self.gs+self.param+self.free])    


















import os
from os.path import isfile

fold="layers/"
fold="layerfiles/"

exit()

imps=[]
clas=[]




with open("layers.py","w") as f:

  for q in os.listdir(fold):
    if not isfile(fold+q):continue
    if "_" in q:continue
    if "makeinit.py" in q:continue
    if not ".py" in q:continue
    
    with open(fold+q,"r") as ff:
      ac=ff.read()

      ac1=ac[:ac.find("\nclass ")]
      ac2=ac[ac.find("\nclass "):]
      
      imps.append(ac1)
      clas.append(ac2)

  i={}
  for imp in imps:
    for lin in imp.split("\n"):
      if not lin in i.keys():i[lin]=1.0

  for ii in i.keys():
    if not "import" in ii:continue
    if len(ii)==0:continue
    if ii[0]=="#":continue
    f.write(ii+"\n")


  for ii in i.keys():
    if "import" in ii:continue
    if len(ii)==0:continue
    if ii[0]=="#":continue
    f.write(ii+"\n")

 
 
  for cl in clas:
    f.write(cl+"\n\n\n\n")







#NOTWORKINGYET
import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gsym(Layer):
  def __init__(self,gs=20,**kwargs):
    self.gs=gs
    super(gsym,self).__init__(**kwargs)

  def get_config(self):
    mi={"gs":self.gs}
    th=super(gsym,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gsym(**config)
  
  def build(self, input_shape):

    self.built=True

  def call(self,x):
    x=x[0]
    xp=K.permute_dimensions(x,(0,2,1))
    
    s=x+xp
    s=K.relu(5*s)-K.relu(5*s-1)    
    
    return s
   

 
  def compute_output_shape(self,input_shape):
    pshape=list(input_shape[0])
    assert len(pshape)==3
    assert pshape[-1]==self.gs
    assert pshape[-2]==self.gs
    return tuple(pshape)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gaddbias(Layer):
  def __init__(self,gs=20,param=40,metrik_init="glorot_uniform",learnable=True,**kwargs):
    self.gs=gs
    self.param=param
    self.metrik_init=metrik_init
    self.learnable=learnable
    super(gaddbias,self).__init__(**kwargs)

  def build(self, input_shape):

    self.bias=self.add_weight(name="bias",
                                shape=(self.gs,self.param),
                                initializer=self.metrik_init,#ignores metrik_init completely
                                trainable=self.learnable,
                                regularizer=None)

#    self.metrik=self.add_weight(name="metrik",
#                                shape=(self.param,1),
#                                initializer=keras.initializers.ones(),#ignores metrik_init completely
#                                trainable=not self.learnable,
#                                regularizer=None)


    self.built=True


  def call(self,x):
    x=x[0]
 
    #print("x",x.shape)
    #print("bias",self.bias.shape)
 
    ret=K.bias_add(x=x,bias=self.bias)
    #print("ret",ret.shape) 

    return ret

    exit()

    
  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    return tuple([input_shape[0],self.gs,self.param])    

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"metrik_init":self.metrik_init,"learnable":self.learnable}
    th=super(gaddbias,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gaddbias(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gpre3(Layer):#just like gpre2, but with flag at 0
  def __init__(self,gs=20,numericC=10000,**kwargs):
    self.gs=gs
    self.numericC=numericC

    super(gpre3,self).__init__(input_shape=(gs,4))#fixed dimension, since specialised layer

  def get_config(self):
    mi={"gs":self.gs,"numericC":self.numericC}
    th=super(gpre3,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gpre3(**config)
  
  def build(self, input_shape):

    self.built=True

  def call(self,x):
    # print("!",x.shape)

    E=K.reshape(x[:,:,0],(-1,self.gs,1))
    p1=K.reshape(x[:,:,1],(-1,self.gs,1))
    p2=K.reshape(x[:,:,2],(-1,self.gs,1))
    p3=K.reshape(x[:,:,3],(-1,self.gs,1))


    
    pt=K.sqrt(p1**2+p2**2)
    p=K.sqrt(pt**2+p3**2)
    iszero=p**2+E**2
    iszero=1-K.relu(1-self.numericC*iszero)+K.relu(-self.numericC*iszero)



    #return iszero

    eta=iszero*0.5*K.log(0.0000000001+(p+p3)/(p-p3+0.0000000001))
    #phi=iszero*t.math.acos(p3/(p+0.0000001))
    phi=iszero*t.math.atan2(p2,p1)

    #print("eta",eta.shape,"phi",phi.shape)

    meta=K.mean(eta,axis=-2)
    mphi=K.mean(phi,axis=-2)

    #print("meta",meta.shape,"mphi",mphi.shape)

    #deta=eta#-meta##not sure if abs here
    #dphi=phi#-mphi##not sure here either


    deta=iszero*K.permute_dimensions(K.permute_dimensions(eta,(1,0,2))-meta,(1,0,2))#not sure if abs here required
    dphi=iszero*K.permute_dimensions(K.permute_dimensions(phi,(1,0,2))-mphi,(1,0,2))#also not sure here either

    #print("deta",deta.shape,"dphi",dphi.shape)

    lpt=iszero*K.log(pt+0.0000001)##
    lE=iszero*K.log(E+0.0000001)##
    
    #print("lpt",lpt.shape,"lE",lE.shape)

    #rpt=K.reshape(pt,(-1,self.gs))

    spt=K.sum(pt,axis=-2)
    #print("spt",spt.shape)
   

    #return K.permute_dimensions(K.log(1.0+K.abs(K.permute_dimensions(pt,(1,0,2))/(K.abs(spt)+1.0))),(1,0,2)) 
    #ispt=1/(spt+0.000000001)
    #print("ispt",ispt.shape)

    ppt=-iszero*K.permute_dimensions(K.log(0.000000001+K.permute_dimensions(pt,(1,0,2))/(spt+0.0000001)),(1,0,2))#please note the added sign in comparison to the original paper
    #ppt=K.reshape(K.permute_dimensions(K.log(0.000000001+K.permute_dimensions(rpt,(1,0,2))/(spt+0.0000001)),(1,0)),(-1,self.gs,1))##
    #ppt=K.reshape(K.permute_dimensions(K.log(0.000000001+K.permute_dimensions(rpt,(1,0))/(spt+0.0000001)),(1,0)),(-1,self.gs,1))##
    #ppt=K.reshape(K.log(0.000000001+K.permute_dimensions(rpt,(1,0))/(spt+0.0000001)),(-1,self.gs,1))##
    #ppt=iszero*K.log(pt/(spt+0.00000001))##
    #print("ppt",ppt.shape)

    sE=K.sum(E,axis=-2)
    #print("sE",sE.shape)
    
    pE=-iszero*K.permute_dimensions(K.log(0.000000001+K.permute_dimensions(E,(1,0,2))/(sE+0.0000001)),(1,0,2))#here was also a sign added
    #pE=-iszero*K.log(sE/(E+0.00000001))##
    #pE=-iszero
    #print("pE",pE.shape)

    dR=K.sqrt(deta**2+dphi**2)##
    #print("dR",dR.shape)

    ret=K.concatenate((iszero,deta,dphi,lpt,lE,ppt,pE,dR),axis=-1)#adding iszero for numerical reasons
   
    #print(ret.shape,x.shape)
    #exit() 

    return ret

    
  def compute_output_shape(self,input_shape):
    shape=list(input_shape)
    assert len(shape)==3
    assert shape[-1]==4
    assert shape[-2]==self.gs
    shape[-1]=8#?
    return tuple(shape)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gbrokengrowth(Layer):
  def __init__(self,inn=20,out=30,param=40,kernel_initializer="glorot_uniform",**kwargs):
    self.inn=inn
    self.out=out
    self.param=param
    self.kernel_initializer=kernel_initializer
    super(gbrokengrowth,self).__init__(**kwargs)

  def get_config(self):
    mi={"inn":self.inn,"out":self.out,"param":self.param}
    th=super(gbrokengrowth,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gbrokengrowth(**config)
  
  def build(self, input_shape):

    self.inc=self.add_weight(name='inc',
                                shape=(self.param*self.inn,self.param*(self.out-self.inn)),
                                initializer=self.kernel_initializer,
                                trainable=True)
    self.built=True

  def call(self,x):
    x=x[0]
    
    f=K.reshape(x,(-1,self.param*self.inn))
    fa=K.dot(f,self.inc)
    r=K.concatenate((f,fa),axis=-1)
    
    return K.reshape(r,(-1,self.out,self.param))
    
     

 
  def compute_output_shape(self,input_shape):
    pshape=list(input_shape[0])
    assert len(pshape)==3
    assert pshape[-1]==self.param
    assert pshape[-2]==self.inn
    pshape[-2]=self.out
    return tuple(pshape)
















import numpy as np
from numpy.random import randint as rndi
from numpy.random import random as rnd

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense,Activation,Flatten,Dropout,Input,Concatenate,Dropout,Reshape,BatchNormalization,Conv2D
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential,Model
from tensorflow.keras.optimizers import Adam,SGD,RMSprop
from tensorflow.keras.utils import plot_model

###import all the "migth be used" special layers

from grapa.layers import *

objects={"gbuilder":gbuilder,"gtbuilder":gtbuilder,"glbuilder":glbuilder,"gtlbuilder":gtlbuilder,"gcutter":gcutter,"gpool":gpool,"gfeat":gfeat,"gl":gl,"gkeepbuilder":gkeepbuilder,"glkeep":glkeep,"gfeatkeep":gfeatkeep,"gkeepcutter":gkeepcutter,"gkeepmatcut":gkeepmatcut,"gtopk":gtopk,"gltk":gltk,"gltknd":gltknd,"gpre1":gpre1,"gpre2":gpre2,"gpre3":gpre3,"gpre4":gpre4,"gpre5":gpre5,"glmlp":glmlp,"gltrivmlp":gltrivmlp,"gadd1":gadd1,"gaddzeros":gaddzeros,"gsym":gsym,"gbrokengrowth":gbrokengrowth,"gpoolgrowth":gpoolgrowth,"gremoveparam":gremoveparam,"gcompool":gcompool,"gcomdepool":gcomdepool,"gvaluation":gvaluation,"ggoparam":ggoparam,"gfromparam":gfromparam,"gaddbias":gaddbias,"gssort":gssort,"glm":glm,"glom":glom,"glim":glim,"ggraphstract":ggraphstract,"gmake1graph":gmake1graph,"gcomdepoolplus":gcomdepoolplus,"gcomdex":gcomdex,"gcomjpool":gcomjpool,"gcomgpool":gcomgpool,"gchooseparam":gchooseparam,"gcomdepoollg":gcomdepoollg,"gcomdiagraph":gcomdiagraph,"gcomreopool":gcomreopool,"gcomparastract":gcomparastract,"glam":glam,"gcomdensemerge":gcomdensemerge,"gcompoolmerge":gcompoolmerge,"gcomgraphcutter":gcomgraphcutter,"gcomdensediverge":gcomdensediverge,"gcomgraphfromparam":gcomgraphfromparam,"gcomgraphcombinations":gcomgraphcombinations,"gcomgraphand":gcomgraphand,"gcomgraphrepeat":gcomgraphrepeat,"gcomgraphlevel":gcomgraphlevel,"gcomgraphand2":gcomgraphand2,"gcomparamlevel":gcomparamlevel,"gliam":gliam,"gcomfullyconnected":gcomfullyconnected,"gcomparamcombinations":gcomparamcombinations,"gcomgraphfrom2param":gcomgraphfrom2param,"gcomextractdiag":gcomextractdiag,"gmultiply":gmultiply,"glcreate":glcreate,"glacreate":glacreate,"gshuffle":gshuffle,"gortho":gortho,"gperm":gperm,"gpartinorm":gpartinorm,"gcutparam":gcutparam,"ghealparam":ghealparam,"gecutter":gecutter}


from grapa.constants import *








def multidense(g,m,q):
  '''runs just a list of Dense Layers (defined by m.mdense* and by the parameter q, which gives width and number of Layers) on the last axis of the input data'''
  for d in q:
    g.X=Dense(d,activation="linear",kernel_initializer=m.mdense_init_kernel,bias_initializer=m.mdense_init_bias,use_bias=m.mdense_usebias)(g.X)
    if m.mdense_batchnorm:g.X=BatchNormalization()(g.X)
    if m.mdense_activation!="linear":g.X=Activation(m.mdense_activation)(g.X)
    g.s.param=d
  return g




def norm(g,scale=True):
  """normalises a network on the last axis, scale decides if there is a learnable multiplicative factor"""
  g.X=BatchNormalization(axis=-1,scale=scale)(g.X)
  return g

def prep(g,m):#preparation handler, return g and input
  """runs my standart preparation on an Input which it defines itself and also returns"""
  inp=Input(shape=(g.s.gs,4))
  feat0=gpre5(gs=g.s.gs)(inp)
  if m.prenorm:norm(g,scale=False)
  g.X=feat0
  g.s.param=4
  return g,inp

def gq(g,m,steps=4):
  """function to work with alternative Input format (here Dense Layer on concat(self_values, neigbour_values)), migth be extended to use Convolutions. Defined by m.gq*"""
  shape0=g.X.shape
  g.X=glcreate(gs=g.s.gs,param=g.s.param)([g.A,g.X])
  for d in denseladder(c=2,n=steps,truestart=True)[::-1]:
    g.X=Dense(int(d*g.s.param),activation="linear",kernel_initializer=m.gq_init_kernel,bias_initializer=m.gq_init_bias,use_bias=m.gq_usebias)(g.X)
    if m.gq_batchnorm:g.X=BatchNormalization()(g.X)
    if m.gq_activation!="linear":g.X=Activation(m.gq_activation)(g.X)
  return g
def gaq(g,m,a,steps=4):
  """like gq but to work on a bit more abstract data (defined by m.gaq*)"""
  shape0=g.X.shape
  g.X=glacreate(gs=g.s.gs,a=a,param=g.s.param)([g.A,g.X])
  for d in denseladder(c=2,n=steps,truestart=True)[::-1]:
    g.X=Dense(int(d*g.s.param),activation="linear",kernel_initializer=m.gqa_init_kernel,bias_initializer=m.gqa_init_bias,use_bias=m.gqa_usebias)(g.X)
    if m.gqa_batchnorm:g.X=BatchNormalization()(g.X)
    if m.gqa_activation!="linear":g.X=Activation(m.gqa_activation)(g.X)
  return g

def gnl(g,m,alin=[],iterations=1,repeat=1,usei=False):
  """a function to just add some graph update functionality without relearning the graph, defined by m.graph*. Can use usei to use inverted Graph update layers instead of the normal ones (to make invertibility easier). Also understands alin (iarities) as a vector"""
  if m.shallcomplex:
    return gq(g,m,steps=m.complexsteps)
  if usei:
    g.X=glim(gs=g.s.gs,param=g.s.param,iterations=iterations,alinearity=alin,self_initializer=m.graph_init_self,neig_initializer=m.graph_init_neig)([g.A,g.X])
  else:
    g.X=glm(gs=g.s.gs,param=g.s.param,iterations=iterations,alinearity=alin,self_initializer=m.graph_init_self,neig_initializer=m.graph_init_self)([g.A,g.X])
  if repeat>1:return gnl(g,alin=alin,iterations=iterations,repeat=repeat-1,usei=usei)
  return g
def learngraph(g,free=0,k=4):
  """just learns a graph (g.A) as a function of the parameters (g.X). Can also add new parameters to g.X (with free) and you can specify how many connections each node should have (k), mainly used by gll"""
  mat,feat=gtopk(gs=g.s.gs,param=g.s.param,free=free,flag=flag,self_interaction=self_interaction,k=k)([g.X])
  g.A=mat
  g.s.param+=free
  g.X=feat
  return g
def gll(g,m,free=0,alin=[],iterations=1,repeat=1,subrepeat=1,usei=False,k=4):
  """gnl + learngraph"""
  g=learngraph(g,free=free,k=k)
  g=gnl(g,m,alin=alin,iterations=iterations,repeat=subrepeat,usei=usei)
  if repeat>1:return gll(g,free=free,alin=alin,iterations=iterations,repeat=repeat-1,usei=usei)
  return g

def ganl(A,X,gs,a,param,m,iterations=1,alin=[],usei=False):
  """gnl but on more abstract graphs, should probably not be used directly unless you unstand what the difference is"""
  if not m.shallacomplex:
    if usei:
      return gliam(gs=gs,a=a,param=param,iterations=iterations,alinearity=alin,self_initializer=m.agraph_init_self,neig_initializer=m.agraph_init_neig)([A,X])
    else:
      return glam(gs=gs,a=a,param=param,iterations=iterations,alinearity=alin,self_initializer=m.agraph_init_self,neig_initializer=m.agraph_init_neig)([A,X])
  else:
    ag=grap(state(gs=gs,param=param))
    ag.X=X
    ag.A=A
    ag=gaq(ag,m,a=a,steps=m.complexasteps)
    return ag.X
def abstr(g,c,alin=[],iterations=1,repeat=1,multiglam=1,pmode="max",gmode="mean"):
  """uses (multiglam) glam to abstract a graph into a factor c smaller graph
  uses pooling to go from c size subgraphs to 1 size dots
  does not chance param at all
  uses (pmode) param pooling mode
  uses (gmode) graph pooling mode"""
  graph,feat1=gcomreopool(gs=g.s.gs,param=g.s.param)([g.A,g.X])#reorder by last param
  graphs=gcomdiagraph(gs=g.s.gs,c=c)([graph])#diagonal graphs
  feats1=gcomparastract(gs=g.s.gs,param=g.s.param,c=c)([feat1])#abstract 3d parameters into 4d ones
  g.s.gs=int(g.s.gs/c)
  for i in range(multiglam):feats1=ganl(graphs,feats1,gs=c,a=g.s.gs,param=g.s.param,iterations=iterations,alin=alin)#run sub graph actions
  feat1=gcompoolmerge(gs=g.s.gs,ags=c,param=g.s.param,mode=pmode)([feats1])  
  graph=gcomgraphcutter(gs=g.s.gs*c,c=c,mode=gmode,cut=cut,c_const=c_const)([graph])#goes from big graph to small graph
  g.A=graph
  g.X=feat1
  if repeat>1:return gabstr(g,c,alin=alin,iterations=iterations,repeat=repeat-1,multiglam=multiglam,pmode=pmode,gmode=gmode)
  return g

def compress(g,m,c,addparam):
  """little brother of abstr, the main difference is, that this does not keep any information of the graph, so you have to retrain it, if you want to do graph actions afterwards"""
  g.X=gcompool(gs=g.s.gs,param=g.s.param,paramo=g.s.param+addparam,c=c,metrik_init=m.compression_init)([g.X])
  g.A=None
  g.s.gs=int(g.s.gs/c)
  g.s.param+=addparam
  return g




def graphatbottleneck(g,m,shallfp=True):
  """handles the bottleneck transformations for a pure graph ae, return g, compressed, new input, shallfp=True=>convert vector in matrix (with gfromparam), can use redense to add a couple dense layers around the bottleneck (defined by m.redense*)"""
  comp=ggoparam(gs=g.s.gs,param=g.s.param)([g.X])
  if m.shallredense:
    for e in m.redenseladder:
      comp=Dense(e,activation=m.redenseactivation,kernel_initializer=m.redenseinit)(comp)
    inn2=Input(m.redenseladder[-1])
    use=inn2
    for i in range(len(m.redenseladder)-1,-1,-1):
      use=Dense(m.redenseladder[i],activation=m.redenseactivation,kernel_initializer=m.redenseinit)(use)
    use=Dense(g.s.gs*g.s.param,activation=m.redenseactivation,kernel_initializer=m.redenseinit)(use)
  else:
    inn2=Input(g.s.gs*g.s.param)
    use=inn2

  if shallfp:
    taef1=gfromparam(gs=g.s.gs,param=g.s.param)([use])
  else:
    taef1=inn2
  g.X=taef1
  g.A=None
  return g,comp,inn2


def denseladder(c,n=3,truestart=False):
  """helper function that generates a list of Dense sizes going from 1 to c in n steps (excluding 1 and c), c can be a list, than returns a list of lists"""
  if type(c)==list:
    ret=[]
    for ac in c:
      ret.append(denseladder(ac,n=n,truestart=truestart))
    return ret

  if truestart:n-=1

  ret=[]
  if truestart:ret.append(1)
  fact=1.0
  mul=c**(1/(n+1))
  for i in range(n):
    fact*=mul
    ret.append(fact)
  return ret

def divtriv(g,c,m,shallgp=True,addDense=[[]],activation=None):
  """trivial graph diverger by a factor of c (does not chance param at all)
  requiregp=True: require ggoparam at the start
  addDense: intermediate Dense Layers, sizes between 1 and c useful"""



  param=g.X
  if shallgp:param=ggoparam(gs=g.s.gs,param=g.s.param)([param])
  for d in addDense[0]:param=Dense(int(d*g.s.gs*g.s.param),activation=activation,kernel_initializer=m.trivial_decompress_init_kernel,bias_initializer=m.trivial_decompress_init_bias)(param)
  g.s.gs*=c[0]#actually not completely the same as magic 29, since there was a param mistake
  param=Dense(g.s.gs*g.s.param,activation,kernel_initializer=m.trivial_decompress_init_kernel,bias_initializer=m.trivial_decompress_init_bias)(param)
  g.X=gfromparam(gs=g.s.gs,param=g.s.param)([param])
  c=c[1:]
  if len(addDense)>0:
    addDense=addDense[1:]
  if len(c)>0:
    return divtriv(g,c,shallgp=False,addDense=addDense)
  return g

def divccll(g,c):
  """easy diverger: diverge by copy"""
  g.X=gmultiply(gs=g.s.gs,param=g.s.param,c=c)([g.X])
  g.A=None
  g.s.gs*=c
  return g





def divpar(g,c,usei=False,alin=[],iterations=1,repeat=1,multiglam=1,amode2="prod"):
  """A parameter like graph diverger by a factor of c (also does not chance param at all)"""

  #print("par on",g.A,g.X,g.s.gs,g.s.param,c)
  #exit()

  #amode2 : and operation modus for graphand2
  park0=gcomfullyconnected(gs=g.s.gs,param=g.s.param)([g.X])#generate initial fully connected graph
  taef1s=gcomdensediverge(gs=g.s.gs,param=g.s.param,paramo=g.s.param,c=c)([g.X])#list of param to list of list of param
  park1=gcomparamcombinations(gs=g.s.gs,param=g.s.param)([g.X])#all possible combinations of parameters
  park2=gcomgraphfrom2param(gs=g.s.gs,param=g.s.param,c=c)([park1])#make each combination into a graph
  park4=gcomgraphlevel(gs=g.s.gs,c=c)([park2])#make matrix of graphs into one graph
  park5=gcomgraphrepeat(gs=g.s.gs,c=c)([park0])#better:scale up park0 by replacing each 1/0 by a 1/0 c*c Matrix
  park6=gcomgraphand2(gs=g.s.gs,c=c,mode=amode2)([park4,park5])#merge the graphs (new graphs, with old graph)
  parkd=gcomextractdiag(gs=g.s.gs,c=c)([park2])#extract the needed matrices for glam from the combination generators

  #if usei:
  #  for i in range(multiglam):taef1s=gliam(gs=c,param=g.s.param,a=g.s.gs,alinearity=alin,iterations=iterations)([parkd,taef1s])#does the subnode graph actions
  #else: 
  #  for i in range(multiglam):taef1s=glam(gs=c,param=g.s.param,a=g.s.gs,alinearity=alin,iterations=iterations)([parkd,taef1s])#does the subnode graph actions
  for i in range(multiglam):taef1s=ganl(parkd,taef1s,gs=c,param=g.s.param,a=g.s.gs,alin=alin,iterations=iterations,usei=usei)

  taef2=gcomparamlevel(gs=g.s.gs,c=c,param=g.s.param)([taef1s])
  g.s.gs*=c
  g.X=taef2
  g.A=park6
  if repeat>1:return divpar(g,c,alin=alin,usei=usei,iterations=iterations,repeat=repeat,multiglam=multiglam,amode2=amode2)
  return g

def divcla(g,c,m,repeat=1):
  """classic graph abstractor, also does not chance the paramsize, just goes from one param to c params, and has one learnable matrix (which is const between the elements). Works by usual parameter divergence, and then by abstracting the graphs, with the constant learnable one"""
  feat0,nmat=gcomdepoolplus(gs=g.s.gs,param=g.s.param,paramo=g.s.param,c=c,metrik_init=keras.initializers.Identity())([g.X])
  rmat=ggraphstract(in1=g.s.gs,in2=c)([g.A,nmat])
  g.A=rmat
  g.X=feat0
  g.s.gs*=c
  if repeat>1:return divcla(g,c,repeat=repeat-1)
  return g
def divcla2(g,c,m,repeat=1):
  """even more simple divcla, the main difference is, that this ignores graphs completely"""
  feat0=gcomdepool(gs=g.s.gs,param=g.s.param,paramo=g.s.param,c=c,metrik_init=keras.initializers.Identity())([g.X])
  g.A=None
  g.X=feat0
  g.s.gs*=c
  if repeat>1:return divcla2(g,c,repeat=repeat-1)
  return g

def divgra(g,c,m,usei=False,alin=[],iterations=1,repeat=1,multiglam=1,amode="prod",amode2="prod"):
  """graph like graph diverger by a factor of c (also does not chance param at all)
  amode  : and operation modus for graphand
  amode2 : and operation modus for graphand2"""
  park0=gcomfullyconnected(gs=g.s.gs,param=g.s.param)([g.X])#generate initial fully connected graph
  taef1s=gcomdensediverge(gs=g.s.gs,param=g.s.param,paramo=g.s.param,c=c)([g.X])#list of param to list of list of param  
  park1=gcomgraphfromparam(gs=g.s.gs,param=g.s.param,c=c)([g.X])#generates graphs from sets of params
  park2=gcomgraphcombinations(gs=g.s.gs,c=c)([park1])#build all combinations of these graphs
  park3=gcomgraphand(gs=g.s.gs,c=c,mode=amode)([park2])#combine graphs
  park4=gcomgraphlevel(gs=g.s.gs,c=c)([park3])#make matrix of graphs into one graph
  park5=gcomgraphrepeat(gs=g.s.gs,c=c)([park0])#better:scale up park0 by replacing each 1/0 by a 1/0 c*c Matrix
  park6=gcomgraphand2(gs=g.s.gs,c=c,mode=amode2)([park4,park5])#merge the graphs (new graphs, with old graph)

  #if usei:
  #  for i in range(multiglam):taef1s=gliam(gs=c,param=g.s.param,a=g.s.gs,alinearity=alin,iterations=iterations)([park1,taef1s])#does the subnode graph actions
  #else:
  for i in range(multiglam):taef1s=ganl(park1,taef1s,gs=c,param=g.s.param,a=g.s.gs,alin=alin,iterations=iterations,usei=usei)
  #  for i in range(multiglam):taef1s=glam(gs=c,param=g.s.param,a=g.s.gs,alinearity=alin,iterations=iterations)([park1,taef1s])#does the subnode graph actions
 
  taef2=gcomparamlevel(gs=g.s.gs,c=c,param=g.s.param)([taef1s])
  g.s.gs*=c
  g.X=taef2
  g.A=park6
  if repeat>1:return divgra(g,c,alin=alin,usei=usei,iterations=iterations,repeat=repeat,multiglam=multiglam,amode=amode,amode2=amode2)
  return g



def remparam(g,nparam):
  """just a simple function to remove overdue parameters"""
  g.X=gremoveparam(gs=g.s.gs,inn=g.s.param,out=nparam)([g.X])
  g.s.param=nparam
  return g


def handlereturn(inn1,raw,com,inn2,decom,shallvae):
  """a nice function to simplify returning values for createbothmodels. Also has some simple size consistency checks
  the variables:
  #inn1  :   initial input Variable
  #raw   :   preconverted input Variable, for comparison sake
  #com   :   compressed Variable
  #inn2  :   input for decoder
  #decom :   decompressed decoder Variable
  #shallvae: shall i thread this like a variational auto encoder? hier just a bodge of an solution"""




  
  if not int(inn1.shape[-1])==4:
    print("returning failed at input1 shape",inn1.shape)
    assert False
  if not int(raw.shape[-1])==int(decom.shape[-1]):
    print("returning failed at comparison shape comparison(1) of",raw.shape,"and",decom.shape)
    assert False
  if len(raw.shape)>2:
    if not int(raw.shape[-2])==int(decom.shape[-2]):
      print("returning failed at comparison shape comparison(2) of",raw.shape,"and",decom.shape)
      assert False
  if not int(com.shape[-1])==int(inn2.shape[-1]):
    print("returning failed at compression shape comparison of",com.shape,"and",inn2.shape)
    assert False

  
  c1=com
  c2=com
  
  if shallvae:
    c1=Dense(int(c1.shape[-1]))(c1)
    c2=Dense(int(c2.shape[-1]))(c2)
    for i in range(100):print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
    print("!!!running variational!!!")
    for i in range(100):print("~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~")
  return inn1,raw,c1,c2,[],inn2,decom

def sortparam(g,m):
  """sorts X by one of its parameters (m.sortindex), just removes the graph"""
  g.X=gssort(gs=g.s.gs,param=g.s.param,index=m.sortindex)([g.X])
  g.A=None
  return g 


def subedge(inp,param,m):
  """one particlenet like update step, uses m.edge*"""
  feat1=Conv2D(param,(1,1),use_bias=m.edgeusebias)(inp)
  feat2=BatchNormalization()(feat1)
  feat3=Activation(m.edgeactivation)(feat2)
  return feat3

def edgeconv(inp,gs,k,param,m):
  """one set of particlenet like update steps, thus use m.edge* like subedge. also similar to gq (here the main difference is the dense vs convolutional structure"""
  feat0=dtcreate(gs=gs,k=k,param=param,flag=0)([inp])
  for i in range(m.edges):feat0=subedge(feat0,param,m=m)
  qfeat=dtdestroy(gs=gs,k=k,param=param)([feat0])
  outp=Activation(m.edgeactivationfinal)(qfeat)
  return outp

def ge(g,m,k=4):
  """the upper level managing particlenet like update steps (like edgeconv and subedge, can mostly use m.edgeconcat to decide if you should concat or replace the output (concat:like particlenet, replace:probably better for autoencoder)"""
  nw=edgeconv(g.X,g.s.gs,k=k,param=g.s.param,m=m)
  if m.edgeconcat:
    g.X=Concatenate(axis=-1)([g.X,nw])
    g.s.param*=2
  else:
    g.X=nw
  g.A=None
  return g


def shuffleinp(g,seed=None):
  """shuffles the inputs, cross particle...sadly does not keep the shuffle constant"""
  q=ggoparam(gs=g.s.gs,param=g.s.param)([g.X])
  
  q=gshuffle(gs=1,param=g.s.param*g.s.gs,seed=seed)([q])

  g.X=gfromparam(gs=g.s.gs,param=g.s.param)([q])

  return g

def orthoinp(g,seed=None):
  """like shuffleinp, but uses an orthogonal matrix instead of shuffle, thus constant, but mixes the inputs in a certain way"""
  q=ggoparam(gs=g.s.gs,param=g.s.param)([g.X])
  
  q=gortho(gs=1,param=g.s.param*g.s.gs,seed=seed)([q])

  g.X=gfromparam(gs=g.s.gs,param=g.s.param)([q])

  return g

def perminp(g):
  """like orthoinp, but uses an permutation matrix instead of an orthogonal one. migth require some improvements in gperm.py before it becomes truly useful"""
  q=ggoparam(gs=g.s.gs,param=g.s.param)([g.X])
  
  q=gperm(gs=1,param=g.s.param*g.s.gs)([q])

  g.X=gfromparam(gs=g.s.gs,param=g.s.param)([q])

  return g

def pnorm(g):
  """runs a normation on each particle and feature, ignoring the first one"""
  zw,g.X=gcutparam(gs=g.s.gs,param1=1,param2=g.s.param-1)([g.X])
  g.X=gpartinorm(gs=g.s.gs,param=g.s.param-1)([g.X])
  g.X=ghealparam(gs=g.s.gs,param1=1,param2=g.s.param-1)([zw,g.X])
  return g
  
def prevcut(g,ops=4):
  """cuts in gs, takes only the last ops values"""
  g.X=gecutter(inn=g.s.gs,param=g.s.param,out=ops)([g.X])
  g.s.gs=ops
  return g

def goparam(g,chanceS=True):
  """transforms the 3d (-1,gs,param) data into 2d (-1,gs*param) ones. You can use chanceS to disallow this function to chance the settings"""
  g.X=ggoparam(gs=g.s.gs,param=g.s.param)([g.X])
  if chanceS:
    g.s.param=g.s.gs*g.s.param
    g.s.gs=1
  return g





def decompress(g,m,c):
  """function to run diverge algorithms on the input. You can choose the diverge algorithm with m.decompress (trivial, paramlike,graphlike,classic,classiclg,ccll), c can be a list (multiple divergences) and also handles the bottleneck actions (define a new input, and return it later). Always returns: g,compressed version,new input"""
  if type(c)!=list:c=[c]
  if m.decompress=="trivial":
    ###trivial magic29 decompression
    g,com,inn2=graphatbottleneck(g,shallfp=False,m=m)
    for ac in c:g=divtriv(g,m=m,c=ac,shallgp=False,addDense=denseladder(c=ac,n=m.trivial_ladder_n),activation=m.trivial_decompress_activation)
    return g,com,inn2
  if m.decompress=="paramlike":
    ###parameter like decompression
    g,com,inn2=graphatbottleneck(g,m=m)
    for ac in c:
      g=divpar(g,c=ac,usei=m.usei,m=m)
      g=gnl(g,usei=m.usei,m=m)
    return g,com,inn2
  if m.decompress=="graphlike":
    ###graph like decompression
    g,com,inn2=graphatbottleneck(g,m=m)
    for ac in c:
      g=divgra(g,c=ac,usei=m.usei,m=m)
      g=gnl(g,usei=m.usei,m=m)
    return g,com,inn2
  if m.decompress=="classic":
    ###classic decompression algorithm, with one learnable constant graph abstraction
    g,com,inn2=graphatbottleneck(g,m=m)
    g.A=gcomfullyconnected(gs=g.s.gs,param=g.s.param)([g.X])
    for ac in c:
      g=divcla(g,c=ac,m=m)
      g=gnl(g,usei=m.usei,m=m)
    return g,com,inn2
  if m.decompress=="classiclg":
    ###classic decompression algorithm, with learnable graph actions
    g,com,inn2=graphatbottleneck(g,m=m)
    for ac in c:
      g=divcla2(g,c=ac,m=m)
      g=gll(g,usei=m.usei,m=m)
    return g,com,inn2
  if m.decompress=="ccll":
    ###copy copy learn learn decompression algorithm
    g,com,inn2=graphatbottleneck(g,m=m)
    for ac in c:
      g=divccll(g,c=ac,m=m)
      g=gll(g,usei=m.usei,m=m)
    return g,com,inn2
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace




#batch_size=100
#epochs=1000
#verbose=2
#lr=0.001



class gmake1graph(Layer):
  def __init__(self,**kwargs):
    super(gmake1graph,self).__init__(**kwargs)

  def get_config(self):
    mi={}
    th=super(gmake1graph,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gmake1graph(**config)
  
  def build(self, input_shape):

    self.built=True

  def call(self,x):
    x=x[0]

    xs=x.shape
    par=1
    for i in range(1,len(xs)):
      par*=xs[i]


    x=K.reshape(x,[-1,par])[:,0]*0+1
    x=K.reshape(x,[-1,1,1])


    return x


 
  def compute_output_shape(self,input_shape):
    shape=list(input_shape[0])
    assert len(shape)>1
    return tuple([shape[0],1,1])
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcomdiagraph(Layer):
  def __init__(self,gs=20,c=2,**kwargs):
    #c=2
    #mode="min"
    assert gs % c==0#assume there is an equal splitting
    self.gs=gs
    self.c=c
    self.ogs=int(self.gs/self.c)
    super(gcomdiagraph,self).__init__(**kwargs)

  def build(self, input_shape):




    self.built=True


  def call(self,x):
    A=x[0]#reordered matrix gs*gs

    #shall take A and take the ogs diagonal c*c matrices, and concat them   

    graphs=[]

    ogs=self.ogs
    c=self.c

    for i in range(ogs):
      start=i*c
      end=(i+1)*c
      amat=A[:,start:end,start:end]
      amat=K.reshape(amat,(-1,1,c,c))
      graphs.append(amat)
    

    ret=K.concatenate(graphs,axis=1)

    return ret


    
  def compute_output_shape(self,input_shape):
    g_shape=input_shape[0]
    assert len(g_shape)==3
    assert g_shape[1]==self.gs
    assert g_shape[2]==self.gs
    return tuple([g_shape[0],self.ogs,self.c,self.c])

  def get_config(self):
    mi={"gs":self.gs,"c":self.c}
    th=super(gcomdiagraph,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomdiagraph(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gshuffle(Layer):#shuffles the param
  def __init__(self,gs=20,param=40,seed=None,**kwargs):
    self.gs=gs
    self.param=param
    self.seed=seed
    super(gshuffle,self).__init__(input_shape=(gs,param))

  def build(self, input_shape):

    self.built=True


  def call(self,x):
    x=x[0]

    x=K.reshape(x,(-1,self.gs,self.param))

    x=K.permute_dimensions(x,(2,0,1))#can only shuffle the first dimension 

    x=t.random.shuffle(x,seed=self.seed)
    
    x=K.permute_dimensions(x,(1,2,0))#dispermute

    return x


  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    if len(input_shape)==2 and self.gs==1:
      assert input_shape[1]==self.param
      return tuple([input_shape[0],self.gs,self.param])
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    return tuple([input_shape[0],self.gs,self.param])    

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"seed":self.seed}
    th=super(gshuffle,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gshuffle(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcompoolmerge(Layer):
  def __init__(self,gs=20,param=30,mode="max",ags=10,**kwargs):
    #c=2
    #mode="min"
    self.gs=gs
    self.param=param
    self.mode=mode
    self.ags=ags
    super(gcompoolmerge,self).__init__(**kwargs)

  def build(self, input_shape):

    #self.trafo=self.add_weight(name="trafo",
    #                           shape=(self.ags*self.param,self.paramo),
    #                           initializer=self.initializer,
    #                           trainable=self.trainable)


    self.built=True


  def call(self,x):
    x=x[0]#params

    #shall take a 4d feature vector of shape (-1,gs,ags,param), and transform it using mode pooling into (-1,gs,paramo)

    print("x",x.shape)

    if self.mode=="max":
      ret=K.max(x,axis=2)
    if self.mode=="min":
      ret=K.min(x,axis=2)
    if self.mode=="mean":
      ret=K.mean(x,axis=2)
    
    #print("ret",ret.shape)

    #exit()

    return ret
    
    




    
  def compute_output_shape(self,input_shape):
    g_shape=input_shape[0]
    assert len(g_shape)==4
    assert g_shape[1]==self.gs
    assert g_shape[2]==self.ags
    assert g_shape[3]==self.param
    return tuple([g_shape[0],self.gs,self.param])

  def get_config(self):
    mi={"gs":self.gs,"ags":self.ags,"param":self.param,"mode":self.mode}
    th=super(gcompoolmerge,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcompoolmerge(**config)
















import numpy as np
import math

from tensorflow.keras import backend as K
from tensorflow.keras.layers import Layer,Dense, Activation
import tensorflow.keras as keras# as k
import tensorflow as t
from tensorflow.keras.models import Sequential
from tensorflow.keras.optimizers import Adam,SGD
from tensorflow.linalg import trace





class gcomdepoolplus(Layer):#simply produces a constant graph additionally to gcomdepool
  def __init__(self,gs=20,param=40,paramo=40,c=2,metrik_init="glorot_uniform",graph_init=keras.initializers.Identity(),learnable=True,**kwargs):
    self.gs=gs
    self.param=param
    self.metrik_init=metrik_init
    self.graph_init=graph_init
    self.c=c
    self.ogs=int(self.gs*self.c)
    self.learnable=learnable
    self.paramo=paramo
    super(gcomdepoolplus,self).__init__(**kwargs)

  def build(self, input_shape):

    self.trafo=self.add_weight(name="trafo",
                                shape=(self.param,self.paramo*self.c),
                                initializer=self.metrik_init,
                                trainable=self.learnable,
                                regularizer=None)

    self.graph=self.add_weight(name="graph",
                                shape=(self.c,self.c),
                                initializer=self.graph_init,
                                trainable=self.learnable,
                                regularizer=None)


    self.built=True


  def call(self,x):
    x=x[0]
 
    #print("x",x.shape)
    

    #print("trafo",self.trafo.shape)


    traf=K.dot(x,self.trafo)
    #print("traf",traf.shape)


    ret=K.reshape(traf,(-1,self.ogs,self.paramo))
    #print("ret",ret.shape)
    
    return ret,self.graph

    #exit()

  def compute_output_shape(self,input_shape):
    input_shape=input_shape[0]
    assert len(input_shape)==3
    assert input_shape[1]==self.gs
    assert input_shape[2]==self.param
    return tuple([input_shape[0],self.ogs,self.paramo]),tuple([self.c,self.c])    

  def get_config(self):
    mi={"gs":self.gs,"param":self.param,"c":self.c,"metrik_init":self.metrik_init,"graph_init":self.graph_init,"learnable":self.learnable,"paramo":self.paramo}
    th=super(gcomdepoolplus,self).get_config()
    th.update(mi)
    return th
  def from_config(config):
    return gcomdepoolplus(**config)













